function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var htm = _interopDefault(require('htm'));

function objectWithoutProperties (obj, exclude) { var target = {}; for (var k in obj) if (Object.prototype.hasOwnProperty.call(obj, k) && exclude.indexOf(k) === -1) target[k] = obj[k]; return target; }
/**
 * @param {Babel} babel
 * @param {object} options
 * @param {string} [options.pragma=h]  JSX/hyperscript pragma.
 * @param {string} [options.tag=html]  The tagged template "tag" function name to process.
 * @param {boolean} [options.monomorphic=false]  Output monomorphic inline objects instead of using String literals.
 * @param {boolean} [options.useBuiltIns=false]  Use the native Object.assign instead of trying to polyfill it.
 * @param {boolean} [options.variableArity=true] If `false`, always passes exactly 3 arguments to the pragma function.
 */

function htmBabelPlugin(ref, options) {
  var t = ref.types;
  if ( options === void 0 ) options = {};

  var pragma = options.pragma === false ? false : dottedIdentifier(options.pragma || 'h');
  var useBuiltIns = options.useBuiltIns;
  var inlineVNodes = options.monomorphic || pragma === false;
  var symbol = Symbol();

  function dottedIdentifier(keypath) {
    var path = keypath.split('.');
    var out;

    for (var i = 0; i < path.length; i++) {
      var ident = propertyName(path[i]);
      out = i === 0 ? ident : t.memberExpression(out, ident);
    }

    return out;
  }

  function patternStringToRegExp(str) {
    var parts = str.split('/').slice(1);
    var end = parts.pop() || '';
    return new RegExp(parts.join('/'), end);
  }

  function propertyName(key) {
    if (key.match(/(^\d|[^a-z0-9_$])/i)) { return t.stringLiteral(key); }
    return t.identifier(key);
  }

  function stringValue(str) {
    if (options.monomorphic) {
      return t.objectExpression([t.objectProperty(propertyName('type'), t.numericLiteral(3)), t.objectProperty(propertyName('tag'), t.nullLiteral()), t.objectProperty(propertyName('props'), t.nullLiteral()), t.objectProperty(propertyName('children'), t.nullLiteral()), t.objectProperty(propertyName('text'), t.stringLiteral(str))]);
    }

    return t.stringLiteral(str);
  }

  function createVNode(tag, props, children) {
    // Never pass children=[[]].
    if (children.elements.length === 1 && t.isArrayExpression(children.elements[0]) && children.elements[0].elements.length === 0) {
      children = children.elements[0];
    }

    if (inlineVNodes) {
      return t.objectExpression([options.monomorphic && t.objectProperty(propertyName('type'), t.numericLiteral(1)), t.objectProperty(propertyName('tag'), tag), t.objectProperty(propertyName('props'), props), t.objectProperty(propertyName('children'), children), options.monomorphic && t.objectProperty(propertyName('text'), t.nullLiteral())].filter(Boolean));
    } // Passing `{variableArity:false}` always produces `h(tag, props, children)` - where `children` is always an Array.
    // Otherwise, the default is `h(tag, props, ...children)`.


    if (options.variableArity !== false) {
      children = children.elements;
    }

    return t.callExpression(pragma, [tag, props].concat(children));
  }

  function flatten(props, result) {
    if ( result === void 0 ) result = [];

    var head = props[symbol];
    var rest = objectWithoutProperties( props, [symbol] );
    var tail = rest;
    if (head) { head.forEach(function (obj) {
      flatten(obj, result);
    }); }

    if (Object.keys(tail).length > 0) {
      result.push(tail);
    }

    return result;
  }

  function spreadNode(args, state) {
    if (args.length > 0 && t.isNode(args[0])) {
      args.unshift({});
    } // 'Object.assign(x)', can be collapsed to 'x'.


    if (args.length === 1) {
      return propsNode(args[0]);
    } // 'Object.assign({}, x)', can be collapsed to 'x'.


    if (args.length === 2 && !t.isNode(args[0]) && Object.keys(args[0]).length === 0) {
      return propsNode(args[1]);
    }

    var helper = useBuiltIns ? dottedIdentifier('Object.assign') : state.addHelper('extends');
    return t.callExpression(helper, args.map(propsNode));
  }

  function propsNode(props) {
    return t.isNode(props) ? props : t.objectExpression(Object.keys(props).map(function (key) {
      var value = props[key];

      if (typeof value === 'string') {
        value = t.stringLiteral(value);
      } else if (typeof value === 'boolean') {
        value = t.booleanLiteral(value);
      }

      return t.objectProperty(propertyName(key), value);
    }));
  }

  function transform(node, state) {
    if (node === undefined) { return t.identifier('undefined'); }
    if (node == null) { return t.nullLiteral(); }
    var tag = node.tag;
    var props = node.props;
    var children = node.children;

    function childMapper(child) {
      if (typeof child === 'string') {
        return stringValue(child);
      }

      return t.isNode(child) ? child : transform(child, state);
    }

    var newTag = typeof tag === 'string' ? t.stringLiteral(tag) : tag;
    var newProps = props ? spreadNode(flatten(props), state) : t.nullLiteral();
    var newChildren = t.arrayExpression(children.map(childMapper));
    return createVNode(newTag, newProps, newChildren);
  }

  function h(tag, props) {
    var children = [], len = arguments.length - 2;
    while ( len-- > 0 ) children[ len ] = arguments[ len + 2 ];

    return {
      tag: tag,
      props: props,
      children: children
    };
  }

  var html = htm.bind(h);

  function treeify(statics, expr) {
    var assign = Object.assign;

    try {
      Object.assign = function () {
        var obj;

        var objs = [], len = arguments.length;
        while ( len-- ) objs[ len ] = arguments[ len ];
        return ( obj = {}, obj[symbol] = objs, obj );
      };

      return html.apply(void 0, [ statics ].concat( expr ));
    } finally {
      Object.assign = assign;
    }
  } // The tagged template tag function name we're looking for.
  // This is static because it's generally assigned via htm.bind(h),
  // which could be imported from elsewhere, making tracking impossible.


  var htmlName = options.tag || 'html';
  return {
    name: 'htm',
    visitor: {
      TaggedTemplateExpression: function TaggedTemplateExpression(path, state) {
        var tag = path.node.tag.name;

        if (htmlName[0] === '/' ? patternStringToRegExp(htmlName).test(tag) : tag === htmlName) {
          var statics = path.node.quasi.quasis.map(function (e) { return e.value.raw; });
          var expr = path.node.quasi.expressions;
          var tree = treeify(statics, expr);
          var node = !Array.isArray(tree) ? transform(tree, state) : t.arrayExpression(tree.map(function (root) { return transform(root, state); }));
          path.replaceWith(node);
        }
      }

    }
  };
}

module.exports = htmBabelPlugin;
